"""Strategies for reading from & writing to config files."""
import logging
import json
import os
from abc import ABCMeta, abstractmethod
from configparser import NoSectionError, RawConfigParser

from .exc import SettingsFileNotFoundError, SettingsFileSectionNotFoundError
from .json import JSONDecoder
from .util import parse_file_name_and_section


__all__ = [
    "Strategy",
    "INIStrategy",
    "INIJSONStrategy",
]


log = logging.getLogger(__name__)


class RawValue(str):

    """Marker for values that couldn't be decoded when reading."""


class Strategy(metaclass=ABCMeta):

    file_types = ()

    @abstractmethod
    def get_defaults(self, file_name):
        """Get default settings from file.

        Returns:
            - Settings from the default section.

        """

    @abstractmethod
    def read_section(self, file_name, section=None):
        """Read settings from specified ``section`` of config file.

        This is where the strategy-specific file-reading logic goes.

        Returns:
            - Settings from the specified section or an empty dict if
              the section isn't present.
            - Whether the section is present.

        """

    @abstractmethod
    def write_settings(self, settings, file_name, section=None):
        """Write settings to file."""

    def parse_file_name_and_section(
        self,
        file_name,
        section=None,
        extender=None,
        extender_section=None,
    ):
        """Parse file name and (maybe) section.

        Delegates to :func:`.util.parse_file_name_and_section` to parse
        the file name and section. If that function doesn't find a
        section, this method should return the default section for the
        strategy via :meth:`get_default_section` (if applicable).

        """
        file_name, section = parse_file_name_and_section(
            file_name,
            section,
            extender,
            extender_section,
        )
        if section is None:
            section = self.get_default_section(file_name)
        return file_name, section

    def read_file(self, file_name, section=None, _finalize=True):
        """Read settings from config file."""
        if _finalize:
            file_name, section = self.parse_file_name_and_section(file_name, section)

        if not os.path.isfile(file_name):
            raise SettingsFileNotFoundError(file_name)

        settings = {}

        result = self._read_one_file(file_name, section)
        file_settings, section_present, extends, extends_section = result

        if extends:
            if extends != file_name:
                result = self.read_file(extends, extends_section, _finalize=False)
                extends_settings, extends_section_present = result
                section_present = section_present or extends_section_present
                settings.update(extends_settings)

        settings.update(file_settings)

        if _finalize:
            if not section_present:
                raise SettingsFileSectionNotFoundError(section)
        else:
            return settings, section_present

        return settings

    def _read_one_file(self, file_name, section, settings=None):
        """Read settings from a single config file."""
        if settings is None:
            settings = self.get_defaults(file_name)

        items, section_present = self.read_section(file_name, section)
        default_extends = settings.get("extends", None)
        extends = items.pop("extends", default_extends)

        if extends:
            result = self.parse_file_name_and_section(
                extends,
                extender=file_name,
                extender_section=section,
            )
            extends, extends_section = result

            if extends == file_name:
                result = self._read_one_file(file_name, extends_section, settings)
                extends_items, extends_section_present, *rest = result
                settings.update(extends_items)
                section_present = section_present or extends_section_present
        else:
            extends_section = None

        settings.update(items)
        settings.pop("extends", None)
        return settings, section_present, extends, extends_section

    def get_default_section(self, file_name):
        return None

    def decode_items(self, items):
        """Bulk decode items read from file to Python objects.

        Args:
            items: A sequence of items as would be returned from
                ``dict.items()``.

        Returns:
            dict: A new dict with item values decoded where possible.
                Values that can't be decoded will be wrapped with
                :class:`RawValue`.

        """
        decoded_items = {}
        for k, v in items:
            try:
                v = self.decode_value(v)
            except ValueError:
                v = RawValue(v)
            decoded_items[k] = v
        return decoded_items

    def decode_value(self, value):
        """Decode value read from file to Python object."""
        return value

    def encode_value(self, value):
        """Encode Python object as value that can be written to file."""
        return value


class LocalSettingsConfigParser(RawConfigParser):
    def options(self, section):
        # Order [DEFAULT] options before section options; the default
        # implementation orders them after.
        options = list(self._defaults.keys())
        try:
            section = self._sections[section]
        except KeyError:
            raise NoSectionError(section) from None
        options.extend(k for k in section.keys() if k not in self._defaults)
        return options

    def optionxform(self, option):
        # Don't alter option names; the default implementation lower
        # cases them.
        return option

    def get_section(self, section):
        """Get section without defaults.

        When retrieving a section from a :class:`RawConfigParser` using
        ``parser[section]``, defaults are included. To get a section's
        items without defaults, we have to access the semi-private
        ``_sections`` instance variable, which is encapsulated here.

        """
        return self._sections[section]


class INIStrategy(Strategy):

    file_types = ("ini",)

    def __init__(self):
        # Cache parsers by file name
        self._parser_cache = {}

    def get_defaults(self, file_name):
        """Get default settings from ``[DEFAULT]`` section of file."""
        parser = self.get_parser(file_name)
        defaults = parser.defaults()
        return self.decode_items(defaults.items())

    def read_section(self, file_name, section):
        parser = self.get_parser(file_name)
        if parser.has_section(section):
            items, section_present = parser.get_section(section), True
        else:
            items, section_present = {}, False
        decoded_items = self.decode_items(items.items())
        return decoded_items, section_present

    def write_settings(self, settings, file_name, section):
        file_name, section = self.parse_file_name_and_section(file_name, section)
        parser = self.make_parser()
        if os.path.exists(file_name):
            with open(file_name) as fp:
                parser.read_file(fp)
        else:
            log.info("Creating new local settings file: %s", file_name)
        if section not in parser:
            log.info("Adding new section to %s: %s", file_name, section)
            parser.add_section(section)
        sorted_keys = sorted(settings.keys())
        for name in sorted_keys:
            value = settings[name]
            parser[section][name] = value
        with open(file_name, "w") as fp:
            parser.write(fp)
        for name in sorted_keys:
            value = settings[name]
            log.info("Saved %s to %s as: %s", name, file_name, value)

    def get_default_section(self, file_name):
        """Returns first non-DEFAULT section; falls back to DEFAULT."""
        if not os.path.isfile(file_name):
            return "DEFAULT"
        parser = self.get_parser(file_name)
        sections = parser.sections()
        section = sections[0] if len(sections) > 0 else "DEFAULT"
        return section

    def get_parser(self, file_name):
        if file_name not in self._parser_cache:
            parser = self.make_parser()
            with open(file_name) as fp:
                parser.read_file(fp)
            self._parser_cache[file_name] = parser
        return self._parser_cache[file_name]

    def make_parser(self):
        return LocalSettingsConfigParser()


class INIJSONStrategy(INIStrategy):

    file_types = ("cfg",)

    def decode_value(self, value):
        value = value.strip()
        if not value:
            return None
        try:
            value = JSONDecoder.loads(value)
        except ValueError:
            value = value.replace("\n", " ")
            raise ValueError(f"Could not parse `{value}` as JSON, number, or datetime")
        return value

    def encode_value(self, value):
        return json.dumps(value)


def get_strategy_types():
    """Get a list of all :class:`Strategy` subclasses.

    The list will be ordered by file type extension.

    """

    def get_subtypes(type_):
        subtypes = type_.__subclasses__()
        for subtype in subtypes:
            subtypes.extend(get_subtypes(subtype))
        return subtypes

    sub_types = get_subtypes(Strategy)
    return sorted(sub_types, key=lambda t: t.file_types[0])


def get_file_type_map():
    """Map file types (extensions) to strategy types."""
    file_type_map = {}
    for strategy_type in get_strategy_types():
        for ext in strategy_type.file_types:
            if ext in file_type_map:
                raise KeyError(
                    f"File type {ext} already registered to " f"{file_type_map[ext]}"
                )
            file_type_map[ext] = strategy_type
    return file_type_map


def guess_strategy_type(file_name_or_ext):
    """Guess strategy type to use for file by extension.

    Args:
        file_name_or_ext: Either a file name with an extension or just
            an extension

    Returns:
        Strategy: Type corresponding to extension or None if there's no
            corresponding strategy type

    """
    if "." not in file_name_or_ext:
        ext = file_name_or_ext
    else:
        name, ext = os.path.splitext(file_name_or_ext)
    ext = ext.lstrip(".")
    file_type_map = get_file_type_map()
    return file_type_map.get(ext, None)



import sys
from collections import Mapping, Sequence

from .base import Base
from .color_printer import color_printer as printer
from .types import EnvSetting, LocalSetting, SecretSetting
from .util import NO_DEFAULT, is_a_tty


class Checker(Base):
    def __init__(
        self,
        file_name,
        section=None,
        registry=None,
        strategy_type=None,
        prompt=None,
    ):
        super().__init__(file_name, section, registry, strategy_type)
        if prompt is None:
            try:
                prompt = is_a_tty(sys.stdin) and is_a_tty(sys.stdout)
            except AttributeError:
                prompt = False
        self.prompt = prompt

    def check(self, obj, prefix=None):
        """Recursively look for :class:`.LocalSetting`s in ``obj``.

        ``obj`` can be a dict, tuple, or list. Other types are skipped.

        This will prompt to get the value of local settings (excluding
        those that have had prompting disabled) that haven't already
        been set locally.

        Returns ``True`` or ``False`` to indicate whether settings were
        successfully checked.

        """
        self._populate_registry(obj, prefix)
        settings_to_write, missing = self._check(obj, prefix, {}, {})
        if settings_to_write:
            self.strategy.write_settings(
                settings_to_write,
                self.file_name,
                self.section,
            )
        if missing:
            for name, local_setting in missing.items():
                if isinstance(local_setting, EnvSetting):
                    message = (
                        f"Env setting `{name}` must be set in settings "
                        f"file or environ ({local_setting.name})"
                    )
                elif isinstance(local_setting, SecretSetting):
                    message = f"Secrete setting `{name}` must be set"
                else:
                    message = f"Local setting `{name}` must be set"
                printer.print_error(message)
            return False
        return True

    def _populate_registry(self, obj, prefix=None):
        if isinstance(obj, Mapping):
            items = obj.items()
        elif isinstance(obj, Sequence) and not isinstance(obj, str):
            items = zip(range(len(obj)), obj)
        else:
            return
        for k, v in items:
            name = k if prefix is None else f"{prefix}.{k}"
            if not isinstance(v, LocalSetting):
                self._populate_registry(v, name)
            else:
                self.registry[v] = name

    def _check(self, obj, prefix, settings_to_write, missing):
        if isinstance(obj, Mapping):
            items = sorted(obj.items(), key=lambda item: item[0])
        elif isinstance(obj, Sequence) and not isinstance(obj, str):
            items = zip(range(len(obj)), obj)
        else:
            return {}, {}

        for k, v in items:
            name = k if not prefix else f"{prefix}.{k}"
            if not isinstance(v, LocalSetting):
                self._check(v, name, settings_to_write, missing)
            elif name not in settings_to_write:
                # Note: If name is already in settings_to_write, this
                # local setting was set as a result of being another
                # setting's default.
                is_set = False
                local_setting = v

                if local_setting.derived_default:
                    # Ensure this setting's default is set if it's also a local setting
                    default_name = self.registry[local_setting.derived_default]
                    if not local_setting.has_default:
                        self._check(
                            {default_name: local_setting.derived_default},
                            None,
                            settings_to_write,
                            missing,
                        )

                if local_setting.prompt:  # prompt for value
                    if self.prompt:
                        v, is_set = self.prompt_for_value(name, v)
                elif local_setting.has_default:  # use default w/o prompting
                    v, is_set = local_setting.default, True
                    v = self.strategy.encode_value(v)
                    msg = f"Using default value `{v!r}` for local " f"setting `{name}`"
                    if local_setting.derived_default:
                        msg += f" (derived from {default_name})"
                    printer.print_warning(msg)

                if is_set:
                    local_setting.value = obj[k] = settings_to_write[name] = v
                else:
                    missing[name] = v

        return settings_to_write, missing

    def prompt_for_value(self, name, local_setting):
        v, is_set = NO_DEFAULT, False
        while not is_set:  # Keep prompting until valid value is set
            printer.print_header(
                f"Enter a value for the local setting `{name}` (as JSON)"
            )
            if local_setting.doc:
                printer.print_header(local_setting.doc)
            if local_setting.has_default:
                msg = f"Hit enter to use default: `{local_setting.default!r}`"
                if local_setting.derived_default:
                    default_name = self.registry[local_setting.derived_default]
                    msg += f" (derived from {default_name})"
                printer.print_header(msg)
            v = input("> ").strip()
            if v:
                try:
                    self.strategy.decode_value(v)
                except ValueError as e:
                    printer.print_error(e)
                else:
                    is_set = local_setting.validate(v)
                    if not is_set:
                        printer.print_error(f"`{v}` is not a valid value for {name}")
            elif local_setting.has_default:
                v, is_set = local_setting.default, True
                v = self.strategy.encode_value(v)
                printer.print_info(f"Using default value for `{name}`")
            else:
                printer.print_error(f"You must enter a value for `{name}`")
        return v, is_set



import json

from .exc import NoDefaultError, NoValueError, DefaultValueError
from .util import NO_DEFAULT


class LocalSetting:

    """Used to mark settings that should be set in the local environment.

    If a default value isn't specified, the local setting must be set.
    Generally, local settings with no default value should not be stored
    in version control.

    If a local setting doesn't have a default value, it will be prompted
    for by default. Pass ``prompt=False`` to disable this (or use
    :class:`.SecretSetting`).

    """

    def __init__(self, default=NO_DEFAULT, prompt=True, doc=None, validator=None):
        self.default = default
        self.derived_default = NO_DEFAULT
        if default is not NO_DEFAULT:
            if isinstance(default, LocalSetting):
                self.derived_default = default
            else:
                if callable(default):
                    default = default()
                try:
                    json.dumps(default)
                except TypeError as exc:
                    raise TypeError(
                        f"{exc}\nDefault value for LocalSetting must "
                        f"be JSON serializable"
                    )
        self.prompt = prompt
        self.doc = doc
        self.validator = validator
        self._value = NO_DEFAULT

    @property
    def has_default(self):
        return self._get_default() is not NO_DEFAULT

    @property
    def has_value(self):
        return self._value is not NO_DEFAULT or self.has_default

    def _get_default(self):
        if self.derived_default:
            default = self.derived_default.value
        else:
            default = self._default
            if callable(default):
                default = default()
        return default

    @property
    def default(self):
        default = self._get_default()
        if default is NO_DEFAULT:
            raise NoDefaultError("Local setting has no default value")
        return default

    @default.setter
    def default(self, default):
        self._default = default

    @property
    def value(self):
        value = self._value
        if value is NO_DEFAULT:
            value = self._get_default()
        if value is NO_DEFAULT:
            raise NoValueError("Local setting has no value")
        return value

    @value.setter
    def value(self, value):
        if not self.validate(value):
            raise ValueError(f"`{value}` is not a valid value")
        self._value = value

    def validate(self, v):
        if self.validator:
            return self.validator(v)
        return True

    def __str__(self):
        class_name = self.__class__.__name__
        try:
            default = repr(self.default)
        except NoDefaultError:
            default = "[NO DEFAULT VALUE]"
        try:
            value = repr(self.value)
        except NoValueError:
            value = "[NO VALUE SET]"
        s = f"<{class_name} with default `{default}` and value `{value}`>"
        return s


class SecretSetting(LocalSetting):

    """Used to mark secret settings.

    Secret settings should never be stored in version control. They will
    always be prompted for if not present in the local settings file.

    Secret settings can have a default generator, which *must* be a
    callable. This is to discourage reuse of the same secret value in
    different environments.

    """

    def __init__(self, default=NO_DEFAULT, doc=None, validator=None):
        if default is not NO_DEFAULT and not callable(default):
            raise DefaultValueError(
                "The default for a secret setting must be a callable; "
                f"got {default!r} of type {type(default)}"
            )
        super().__init__(default, True, doc, validator)


class EnvSetting(LocalSetting):

    """Setting fetched from environ.

    Env settings can have a default set in the local settings file,
    which will be overridden if the specified environment variable is
    set.

    Env settings can also be used standalone without a local settings
    file.

    Example::

        # settings.py
        #
        # This setting will be pulled from the ENV_SETTING environment
        # variable.
        ENV_SETTING = EnvSetting("ENV_SETTING")

    Env settings can be nested too::

        # settings.py
        DATABASES = {
            "default": {
                "USER": EnvSetting("DATABASE_USER"),
                "PASSWORD": EnvSetting("DATABASE_PASSWORD"),
                "NAME": EnvSetting("DATABASE_NAME"),
            }
        }

    """

    def __init__(self, name, doc=None, validator=None):
        super().__init__(NO_DEFAULT, False, doc, validator)
        self.name = name



from __future__ import print_function

import sys

from .util import is_a_tty


class ColorPrinterMeta(type):
    def __init__(cls, *args, **kwargs):
        def _make_methods(_color):
            def _print(self, *a, **kw):
                kw["color"] = _color
                return self.print(*a, **kw)

            def _string(self, *a, **kw):
                return self.string(_color, *a, **kw)

            setattr(cls, f"print_{color}", _print)
            setattr(cls, f"string_{color}", _string)

        for color in cls.colors:
            _make_methods(color)


class ColorPrinter(metaclass=ColorPrinterMeta):

    """Prints things in color (or not).

    Default colors can be overridden by passing a dict of colors to
    the constructor.

    Use stand-alone or as a mixin::

        >>> printer = ColorPrinter()
        >>> printer.print('boring old message')
        boring old message
        >>> printer.string('none', 'boring old message')
        'boring old message\\x1b[0m'
        >>> printer.print_info('check this out')
        check this out
        >>> printer.print_error('whoopsie')
        whoopsie
        >>> printer.string('error', 'whoopsie')
        '\\x1b[91mwhoopsie\\x1b[0m'
        >>> MyClass = type('MyClass', (ColorPrinter,), {})
        >>> my_obj = MyClass(colors={'header': '\033[96m'})
        >>> my_obj.print_header('Header')
        Header
        >>> my_obj.string('header', 'Header')
        '\\x1b[96mHeader\\x1b[0m'

    Note: This uses the print function from Python 3.

    """

    colors = {
        "header": "\033[95m",
        "info": "\033[94m",
        "success": "\033[92m",
        "warning": "\033[93m",
        "error": "\033[91m",
        "reset": "\033[0m",
        "none": "",
    }

    def __init__(self, colors=None):
        if colors is not None:
            self.colors = self.colors.copy()
            self.colors.update(colors)

    def print(self, *args, **kwargs):
        """Like built-in ``print()`` but colorizes strings.

        Pass ``color`` as a keyword arg to colorize ``*args`` before
        printing them. If no ``color`` is passed, *args will printed
        without color.

        """
        color = kwargs.pop("color", None)
        file = kwargs.get("file", sys.stdout)
        if color and is_a_tty(file):
            string = self.string(color, *args, **kwargs)
            print(string, **kwargs)
        else:
            print(*args, **kwargs)

    def string(self, color, *args, **kwargs):
        """Returns a colorized string (joining ``args`` into one str).

        The arguments for this are similar to the built-in ``print()``.
        ``sep`` is a space by default, but ``end`` is an empty string.

        """
        color = self.colors[color]
        sep = kwargs.get("sep", " ")
        end = kwargs.get("end", "")
        string = sep.join(str(a) for a in args)
        string = f"{color}{string}{self.colors['reset']}"
        if end:
            string += end
        return string


color_printer = ColorPrinter()



import inspect
import os
import sys
from json import loads

from .color_printer import color_printer as printer
from .exc import LocalSettingsError, SettingsFileDidNotPassCheck
from .loader import Loader
from .util import abs_path, get_file_name, load_dotenv

# Exported (but unused locally)
from .checker import Checker  # noqa: exported
from .exc import SettingsFileNotFoundError  # noqa: exported
from .settings import Settings  # noqa: exported
from .types import EnvSetting, LocalSetting, SecretSetting  # noqa: exported
from .util import NO_DEFAULT  # noqa: exported
from .util import get_default_file_names
from .__main__ import make_local_settings  # noqa: exported


__version__ = "2.0a5"


def load_and_check_settings(
    base_settings,
    file_name=None,
    section=None,
    base_path=None,
    strategy_type=None,
    disable=None,
    prompt=None,
    quiet=None,
    env_only=False,
    dotenv_file=None,
    dotenv_file_name=".env",
):
    """Merge local settings from file with base settings, then check.

    Returns a new dict containing the base settings and the loaded
    settings. Includes:

        - base settings
        - settings from extended file(s), if any
        - settings from file
        - settings from env, if any env settings are defined

    Settings loaded from the local settings file(s) will override base
    settings. If any env settings are defined and set in the environ,
    they will override base settings and settings loaded from file(s).

    After all settings are loaded from the various sources, they will be
    checked to ensure that all local settings have been set.

    If a file name is passed and the file exists, local settings will be
    loaded from it and, if ``prompt=True`` (see below), any missing
    settings will be appended to it. If the file doesn't exist, it will
    be created and, if ``prompt=True``, all settings will be added to
    it.

    If a file name isn't passed: if the ``LOCAL_SETTINGS_FILE``
    environment variable is set, the specified file will be used;
    otherwise ``{base_path}/local.{ext}`` will be used. Settings will be
    appended or added as described above.

    ``base_path`` is used when ``file_name`` is relative; if it's not
    passed, it will be set to the current working directory.

    When ``prompt`` is ``True``, the user will be prompted for missing
    local settings, which will be added to the specified or discovered
    local settings file. By default, the user is prompted only when
    running on TTY. The ``LOCAL_SETTINGS_CONFIG_PROMPT`` environment
    variable can be used to set ``prompt``.

    When ``quiet`` is ``True``, informational messages will not be
    printed. The ``LOCAL_SETTINGS_CONFIG_QUIET`` environment variable
    can be used to set ``quiet``.

    When ``env_only`` is ``True``, only env settings will be loaded--
    all the file loading steps noted above will be skipped. Note that
    if other, non-env local settings are defined, this will cause an
    error when ``env_only=True``.

    ``dotenv_file`` can be used to specify a .env file for loading
    local settings configuration environment variables. This can be the
    same .env file used for local settings or a different one. It can be
    specified as an absolute, relative (to ``base_path``), or asset
    path.

    .. note:: When setting flags via environment variables, use a JSON
        value like 'true', '1', 'false', or '0'.

    See :meth:`.Loader.load` and :meth:`.Checker.check` for more info.

    """
    environ_config = get_config_from_environ(dotenv_file, base_path, dotenv_file_name)
    disable = environ_config["disable"] if disable is None else disable
    if disable:
        return {}
    prompt = get_prompt(environ_config, prompt, env_only)
    quiet = environ_config["quiet"] if quiet is None else quiet
    file_name = get_local_settings_file_name(file_name, base_path, env_only)
    try:
        loader = Loader(file_name, section, strategy_type=strategy_type)
        settings, success = loader.load_and_check(base_settings, prompt)
    except KeyboardInterrupt:
        # Loading/checking of local settings was aborted with Ctrl-C
        # during prompting. This isn't an error, but we don't want to
        # continue.
        if not quiet:
            printer.print_warning("\nAborted loading/checking of local settings")
        # XXX: It seems wonky sys.exit() here; seems like the exception
        #      should just be re-raised.
        sys.exit(0)
    if loader.section:
        file_name = f"{loader.file_name}#{loader.section}"
    else:
        file_name = loader.file_name
    if not success:
        raise SettingsFileDidNotPassCheck(file_name)
    if not quiet:
        printer.print_success(f"Settings loaded successfully from {file_name}")
    return settings


def get_prompt(environ_config, prompt, env_only):
    if env_only:
        if prompt:
            raise LocalSettingsError("The prompt flag can't be set when env_only=True")
        prompt = False
    else:
        prompt = environ_config["prompt"] if prompt is None else prompt
    return prompt


def get_local_settings_file_name(file_name, base_path, env_only):
    if env_only:
        if file_name:
            raise LocalSettingsError(
                "A local settings file can't be specified when env_only=True"
            )
        file_name = None
    else:
        if file_name is None:
            file_name = get_file_name()
        if file_name is None:
            cwd = os.getcwd()
            default_file_names = ", ".join(get_default_file_names())
            raise SettingsFileNotFoundError(
                f"No local settings file was specified and no default "
                f"settings file was found in the current working directory "
                f"(cwd = {cwd}, defaults = {default_file_names})"
            )
        base_path = base_path or os.getcwd()
        file_name = abs_path(file_name, relative_to=base_path)
    return file_name


def inject_settings(base_settings=None, **kwargs):
    """Inject local settings into settings module.

    Call this from the global scope of your Django settings module to
    load settings from the local settings file and inject them into the
    settings module::

        from local_settings import inject_settings
        inject_settings()

    This is equivalent to, but much less tedious than, the following::

        from local_settings import load_and_check_settings
        globals.update(load_and_check_settings(globals()))

    """
    if base_settings is None:
        frame = inspect.stack()[1][0]
        base_settings = frame.f_globals
    settings = load_and_check_settings(base_settings, **kwargs)
    base_settings.update(settings)


def get_config_from_environ(dotenv_file, base_path, file_name=".env"):
    def get(name, default="null"):
        name = name.upper()
        name = f"LOCAL_SETTINGS_CONFIG_{name}"
        return loads(os.environ.get(name, default))

    load_dotenv(dotenv_file, base_path, file_name)
    options = (
        ("disable", "false"),
        ("prompt", "null"),
        ("quiet", "false"),
    )
    return {n: get(n, default) for (n, default) in options}



import json

import arrow


class JSONDecoder(json.JSONDecoder):

    """Custom JSON decoder.

    Handles:

    - Trailing commas
    - Binary, octal, and hex values
    - Unary + prefix on ints and floats
    - Underscore separators in numbers
    - Dates & times

    .. note:: Number, date, and time conversions currently only work
        for *non-nested* values. The JSON decoder in the Python standard
        libary isn't easy, or perhaps even possible at all, to convert
        nested values.

    Examples::

        >>> loads = JSONDecoder.loads

        >>> loads('["a", "b", 1,]')
        ['a', 'b', 1]
        >>> loads('""')
        ''
        >>> loads('1')
        1
        >>> loads('"abc"')
        'abc'
        >>> loads("[1, 2, 3]")
        [1, 2, 3]
        >>> loads("[1, 2, 3,]")
        [1, 2, 3]
        >>> loads('{"a": 1, "b": 2}')
        {'a': 1, 'b': 2}
        >>> loads('{"a": 1, "b": 2,}')
        {'a': 1, 'b': 2}
        >>> loads('["\\\\"a", "b",]')
        ['"a', 'b']

        >>> loads("0b11")
        3
        >>> loads("-0b11")
        -3
        >>> loads("0b_1_1")
        3
        >>> loads("0o11")
        9
        >>> loads("-0o11")
        -9
        >>> loads("0x11")
        17
        >>> loads("-0x11")
        -17
        >>> loads("+1")
        1
        >>> loads("-1")
        -1
        >>> loads("1_000_000")
        1000000

        >>> loads("+1e3")
        1000.0
        >>> loads("-1e3")
        -1000.0
        >>> loads("1e-3")
        0.001
        >>> loads("1_000.0")
        1000.0

        >>> d = loads("2021-06-23")
        >>> (d.year, d.month, d.day, d.hour, d.minute) == (2021, 6, 23, 0, 0)
        True

        >>> d = loads("2021-06-23T12:30")
        >>> (d.year, d.month, d.day, d.hour, d.minute) == (2021, 6, 23, 12, 30)
        True

        >>> t = arrow.now()
        >>> d = loads("12:30:00.000000")
        >>> (d.year, d.month, d.day, d.hour, d.minute) == (t.year, t.month, t.day, 12, 30)
        True


    """

    @classmethod
    def loads(cls, s, **kwargs):
        return json.loads(s, cls=cls, **kwargs)

    def decode(self, s):
        try:
            return super().decode(s)
        except ValueError:
            try:
                return self._try_alternative_decodings(s)
            except ValueError:
                message = f"Could not convert string to JSON: {s}"
                raise ValueError(message) from None

    def _try_alternative_decodings(self, s):
        """Try alternative decodings.

        This is used only when the original value wasn't parsed
        successfully as JSON. The following strategies are attempted:

        - If the value appears to be a JSON list or dict, trailing
          commas will be removed and the object will be decoded again;
          if the second attempt fails, the error is propagated
        - The value will be converted to an int if possible by calling
          `int(value, <base>)`
        - The value will be converted to a float if possible by calling
          `float(value)`
        - The value will be converted to a datetime if possible

        .. note:: Currently, the last three strategies only work for
            non-nested values.

        """
        if s[0] in "{[":
            s = self._remove_trailing_commas(s)
            return super().decode(s)

        try:
            return self._try_int(s)
        except ValueError:
            pass

        try:
            return float(s)
        except ValueError:
            pass

        try:
            return arrow.get(s).datetime
        except ValueError:
            pass

        return self._try_time(s)

    def _try_time(self, s):
        """Convert time to a date/time on today's date."""
        formats = ["HH:mm", "HH:mm:ss", "HH:mm:ss.S"]
        for f in formats:
            try:
                time = arrow.get(s, f)
            except ValueError:
                pass
            else:
                value = arrow.now().floor("day")
                value = value.replace(
                    hour=time.hour,
                    minute=time.minute,
                    second=time.second,
                )
                return value.datetime
        raise ValueError(f"Could not parse value as time: {s}")

    def _remove_trailing_commas(self, s):
        """Remove trailing commas from all JSON object and arrays."""
        stack = []
        collector = []
        in_string = False
        pairs = {"{": "}", "[": "]"}
        opening = None
        closing = None

        # Previous non-whitespace index & char
        prev = None, None

        for i, c in enumerate(s):
            if c == '"' and prev[1] != "\\":
                in_string = not in_string
            elif not in_string:
                if c.isspace():
                    collector.append(c)
                    continue
                elif c in pairs:
                    opening = c
                    closing = pairs[c]
                    stack.append(opening)
                elif c == closing:
                    if stack and stack[-1] == opening:
                        stack.pop()
                        if stack:
                            opening = stack[-1]
                            closing = pairs[opening]
                        else:
                            opening = None
                            closing = None
                        prev_i, prev_c = prev
                        if prev_c == ",":
                            collector[prev_i] = None
                    else:
                        raise ValueError(f"Mismatched {closing}")

            collector.append(c)
            prev = i, c

        return "".join(c for c in collector if c)

    def _try_int(self, s, bases={"0b": 2, "0o": 8, "0x": 16}):
        """Try converting value to int.

        Handles binary, octal, and hex values. Also handles + and -
        prefixes (JSON only handles -).

        """
        s = s.lower()
        if s[0] in "+-":
            prefix = s[0]
            s = s[1:]
        else:
            prefix = ""
        base = bases.get(s[:2]) or 10
        return int(f"{prefix}{s}", base)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['local_settings']

package_data = \
{'': ['*']}

install_requires = \
['Django>=2.2', 'arrow>=1.1.0,<2.0.0', 'python-dotenv>=0.17.1,<0.18.0']

entry_points = \
{'console_scripts': ['make-local-settings = '
                     'local_settings:make_local_settings']}

setup_kwargs = {
    'name': 'django-local-settings',
    'version': '2.0a5',
    'description': 'Define Django settings in config files and/or via environment variables',
    'long_description': '# Local settings for Django projects\n\n## New in version 2\n\n- Settings can now be defined as environment variables (env settings)\n- The values for env settings can be specified in a `.env` file and\n  they\'ll be loaded automatically, or any standard method of setting\n  environment variables can be used\n- Secret settings can now have a callable default (they weren\'t allowed\n  to have a default before)\n- List and dict settings can now have trailing commas\n- Literal date & time values can now be specified (`2021-06-23`, etc)\n- Additional number syntaxes are now supported (any valid Python `int`\n  or `float`)\n- Support for Python 3.5 and below was dropped; supporting only 3.6 and\n  up is much simpler\n- Support for old, unsupported versions of Django was dropped\n\nThis package attempts to solve the problem of handling local settings in\nDjango projects. Local settings by definition can\'t be pre-defined,\nalthough perhaps they can have a reasonable default value (mainly useful\nfor development). Another class of local settings are *secret* settings;\nthese definitely shouldn\'t be pre-defined and should never be added to\nversion control.\n\nThe problems with local settings are:\n\n- How to specify which settings are local\n- How to inform others which settings are local (or secret)\n- How to actually give the local settings a value\n- How to verify that local settings have been given a valid value\n- How to ensure new local settings get set\n- How to ensure local (and esp. secret) settings don\'t get added to\n  version control\n\nOne common approach is to create a local settings template module with\ndummy/default values. When new developers start working on a project,\nthey copy this file (e.g., `local_settings.py.template =>\nlocal_settings.py`), which is typically excluded from version control.\nThis approach at least identifies which settings are local, but it\'s not\nvery convenient with regard to setting values and ensuring those values\nare valid. Also, instead of giving you a friendly heads-up when you\nforget to set a local setting, it barfs out an exception.\n\nThis package takes the approach that there will be only one settings\nmodule* per project in the standard location: `{project}.settings`. That\nmodule defines/overrides Django\'s base settings in the usual way *plus*\nit defines which settings are local and which are secret.\n\nIn addition to the settings module, there will be one or more settings\n*files*. These are standard INI files with the added twist that the\nvalues are JSON encoded. The reasoning behind this is to use a simple,\nstandard config file format while still allowing for easy handling of\nnon-string settings.\n\nIn addition, interpolation is supported using Django-style `{{ ... }}`\nsyntax. This can be handy to avoid repetition.\n\nOnce the local settings are defined, *any missing settings will be\nprompted for in the console* (with pretty colors and readline support).\n\nIn addition to settings files, settings can defined via environment\nvariables. These can be defined in a `.env` file or using any other\nmechanism for setting environment variables. When using a `.env` file,\nthe values will be read in automatically; when *not* using a `.env`\nfile, the corresponding environment variables will need to be set prior\nto loading the local settings.\n\nEnv settings are typically strings like passwords and API tokens, but\nthey will proccessed like other settings--values will be loaded as JSON\nand interpolated, etc.\n\n## Features\n\n- Local settings can be defined in a settings file (or files) *or* as\n  environment variables (AKA env settings)\n- When using env settings, environment variables will be loaded from\n  a `.env` file automatically, if present\n- Missing local settings will be prompted for (only when running on a\n  TTY/console)\n- Local settings can be defined with validators\n- Local settings can be defined with doc strings\n- Local settings can be nested in settings lists and dicts\n- Settings files can extend from each other\n- Settings values can be injected into other settings values using a\n  special syntax (AKA interpolation, similar to the standard library\'s\n  `configparser`)\n- Includes a script to easily generate local settings files for\n  different environments\n- Supports Python 3.6 - 3.9\n- Supports Django 2.2 - 3.2\n\n## Basic usage\n\n- In your project\'s settings *module*, import the `inject_settings`\n  function along with the types of settings you need:\n\n        from local_settings import (\n            inject_settings,\n            LocalSetting,\n            SecretSetting,\n        )\n\n- Then define some base settings and local settings:\n\n        # project/settings.py\n        from django.core.management import utils\n\n        # This is used to demonstrate interpolation.\n        PACKAGE = "local_settings"\n  \n        DEBUG = LocalSetting(default=False)\n  \n        # This setting will be loaded from the environment variable\n        # API_TOKEN, which can be defined in a .env file or set directly\n        # in the environment.\n        SOME_SERVICE = {\n            "api_token": EnvSetting("API_TOKEN"),\n        }\n  \n        DATABASES = {\n            "default": {\n                "ENGINE": "django.db.backends.postgresql",\n                "NAME": LocalSetting(default="{{ PACKAGE }}"),\n                "USER": LocalSetting(""),\n                "PASSWORD": SecretSetting(),\n                "HOST": LocalSetting(""),\n                "PORT": "",\n            },\n        }\n  \n        # If a secret setting specifies a default, it must be a callable\n        # that generates the value; this discourages using the same\n        # secret in different environments.\n        SECRET_KEY = SecretSetting(\n            default=utils.get_random_secret_key,\n            doc="The secret key for doing secret stuff",\n        )\n\n    Local settings can be nested inside other settings. They can also\n    have doc strings, which are displayed when prompting, and default\n    values or value generators, which are used as suggestions when\n    prompting.\n\n    This also demonstrates interpolation. The `DATABASES.default.NAME`\n    setting will be replaced with the `PACKAGE` setting, so that its\n    default value is effectively `\'top_level_package\'`.\n\n- *After* all the local settings are defined, add the following line:\n\n        inject_settings()\n\n    This initializes the local settings loader with the base settings\n    from the settings *module* and tells it which settings are local\n    settings. It then merges in the settings from the settings *file*.\n\n    `inject_settings()` loads the project\'s local settings from a file\n    (`$CWD/local.cfg` by default), prompting for any that are missing,\n    and/or environment variables, and returns a new dictionary with\n    local settings merged over any base settings. When not running on\n    a TTY/console, missing local settings will cause an exception to be\n    raised.\n\n    After `inject_settings()` runs, you\'ll be able to access the local\n    settings in the settings module as usual, in case some dynamic\n    configuration is required. For example, you could do `if DEBUG:\n    ...`. At this point, `DEBUG` is no longer a `LocalSetting`\n    instance--it\'s a regular Python `bool`.\n\n- Now you can run any `manage.py` command, and you will be prompted to\n  enter any missing local settings. On the first run, the settings file\n  will be created. On subsequent runs when new local settings are added\n  to the settings module, the settings file will be appended to.\n\n- Alternatively, you can run the included `make-local-settings` script\n  to generate a local settings file.\n\n## Advanced usage\n\nTODO: Discuss using multiple settings files, extending a settings file\nfrom another file, how to specify a settings file other than the default\nof `local.cfg`, editing settings files directly, &c.\n',
    'author': 'Wyatt Baldwin',
    'author_email': 'self@wyattbaldwin.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/wylee/django-local-settings',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
